# -*- coding: utf-8 -*-
"""
biosppy.version
---------------

Version tracker.

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

VERSION = (0, 8, 0)
__version__ = ".".join(map(str, VERSION))
