# -*- coding: utf-8 -*-
"""
biosppy.signals
---------------

This package provides methods to synthesize common
physiological signals (biosignals):
    * Electrocardiogram (ECG)

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# compat
from __future__ import absolute_import, division, print_function

# allow lazy loading
from . import ecg
